import requests
from requests import Session
import os
from typing import Optional, Tuple, Union
import configparser
import numpy as np
from numbers import Number
from getpass import getpass
from pathlib import Path
from PIL import Image
from io import BytesIO


def create_json_session(user: Optional[str] = None,
                        password: Optional[str] = None,
                        web_host: Optional[str] = None,
                        verify: Optional[bool] = True,
                        server_name: Optional[str] = 'omero',
                        config_path: Optional[str] = None
                        ) -> Tuple[dict, Optional[Session], Optional[str]]:
    """Create an OMERO connection using the JSON API

    This function will create an OMERO connection by populating certain
    parameters for a request using the ``requests`` library by the
    procedure described in the notes below. Note that this function may
    ask for user input, so be cautious if using in the context of a script.

    Parameters
    ----------
    user : str, optional
        OMERO username.

    password : str, optional
        OMERO password.

    web_host : str, optional
        OMERO.web host.

    verify : boolean, optional
        Whether to verify SSL certificates when making requests.

    server_name : str, optional
        OMERO server name as configured in your OMERO.web installation. For
        a "default" OMERO.web installation, it will normally be 'omero'.

    config_path : str, optional
        Path to directory containing '.ezomero' file that stores connection
        information. If left as ``None``, defaults to the home directory as
        determined by Python's ``pathlib``.

    Returns
    -------
    login_rsp : JSON object
        JSON containing the response to the ``POST`` request sent for log in.

    session : ``requests`` Session object or None
        The effective ``requests`` session that will be used for further
        requests to the JSON API.

    base_url : str or None
        Base URL for further requests, retrieved via JSON API request

    Notes
    -----
    The procedure for choosing parameters for initializing a JSON API session
    is as follows:

    1) Any parameters given to `create_json_connection` will be used to
       initialize a JSON session

    2) If a parameter is not given to this function, populate from variables
       in ``os.environ``:
        * OMERO_USER
        * OMERO_PASS
        * OMERO_WEB_HOST

    3) If environment variables are not set, try to load from a config file.
       This file should be called '.ezomero'. By default, this function will
       look in the home directory, but ``config_path`` can be used to specify
       a directory in which to look for '.ezomero'.

       The function ``ezomero.store_connection_params`` can be used to create
       the '.ezomero' file.

       Note that passwords can not be loaded from the '.ezomero' file. This is
       to discourage storing credentials in a file as cleartext.

    4) If any remaining parameters have not been set by the above steps, the
       user is prompted to enter a value for each unset parameter.

    Note that, by default, this function will use the latest API version
    available, in case multiple versions are offered by the OMERO.web server.
    """
    # load from .ezomero config file if it exists
    if config_path is None:
        config_fp = Path.home() / '.ezomero'
    elif type(config_path) is str:
        config_fp = Path(config_path) / '.ezomero'
    else:
        raise TypeError('config_path must be a string')

    config_dict: Union[None, configparser.SectionProxy] = None
    if config_fp.exists():
        config = configparser.ConfigParser()
        with config_fp.open() as fp:
            config.read_file(fp)
        try:
            config_dict = config["JSON"]
        except KeyError:
            raise KeyError('.ezomero does not contain JSON information.')

    # set user
    if user is None:
        if config_dict:
            user = config_dict.get("OMERO_USER", user)
        user = os.environ.get("OMERO_USER", user)
    if user is None:
        user = input('Enter username: ')

    # set password
    if password is None:
        password = os.environ.get("OMERO_PASS", password)
    if password is None:
        password = getpass('Enter password: ')

    # set web host
    if web_host is None:
        if config_dict:
            web_host = config_dict.get("OMERO_WEB_HOST", web_host)
        web_host = os.environ.get("OMERO_WEB_HOST", web_host)
    if web_host is None:
        web_host = input('Enter host: ')

    session = requests.Session()
    # Start by getting supported versions from the base url...
    api_url = '%s/api/' % web_host
    r = session.get(api_url, verify=verify)
    # we get a list of versions
    versions = r.json()['data']
    # use most recent version...
    version = versions[-1]
    # get the 'base' url
    base_url = version['url:base']
    r = session.get(base_url)
    # which lists a bunch of urls as starting points
    urls = r.json()
    servers_url = urls['url:servers']
    login_url = urls['url:login']

    # To login we need to get CSRF token
    token_url = urls['url:token']
    token = session.get(token_url).json()['data']
    # We add this to our session header
    # Needed for all POST, PUT, DELETE requests
    session.headers.update({'X-CSRFToken': token,
                            'Referer': login_url})

    # List the servers available to connect to
    servers = session.get(servers_url).json()['data']
    servers = [s for s in servers if s['server'] == server_name]
    if len(servers) < 1:
        raise Exception("Found no server called '%s'" % server_name)
    server = servers[0]

    # Login with username, password and token
    payload = {'username': user,
               'password': password,
               # Using CSRFToken in header
               'server': server['id']}

    r = session.post(login_url, data=payload)
    login_rsp = r.json()
    try:
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        # Whoops it wasn't a 200
        print("Error {}".format(e))
        raise
    assert login_rsp['success']

    # Can get our 'default' group

    return login_rsp, session, base_url


def get_rendered_jpeg(session: Session, base_url: str, img_id: int,
                      scale: Union[float, int]) -> np.ndarray:
    """Get a numpy array from a rendered JPEG at given scale factor
    of an image.

    Parameters
    ----------
    session : ``requests`` Session object
        ``requests`` session that has been initialized with login parameters.
    base_url : str
        Base URL for further requests, retrieved via JSON API request
    img_id : int
        ID of ``Image``.
    scale : float or int
        Scaling factor for the returned JPEG. ``1`` returns original size,
        ``2`` scales each dimension down by half, and so on.

    Returns
    -------
    pixels : ndarray
        Array containing pixel values from rendered JPEG of the OMERO image.

    Examples
    --------
    # Get a JPEG of an entire image as a numpy array:

    >>> jpeg_array = get_rendered_jpeg(session, base_url, 314, 1)

    """
    if not isinstance(session, requests.sessions.Session):
        raise TypeError('session must is a Requests session object')
    if type(base_url) is not str:
        raise TypeError('base_url must be a string')
    if type(img_id) is not int:
        raise TypeError('img_id must be an int')
    if not isinstance(scale, Number):
        raise TypeError('scale must be a number')
    # magical code for correct address from the json api session and image id
    r = session.get(base_url)
    try:
        r.raise_for_status()
    except requests.exceptions.ConnectionError as e:
        # Whoops it wasn't a 200
        print("Error {!r}: received response {!r} with content: \
                         {!r}".format(e, r.status_code, r.content))
        raise
    host = base_url.split("/api")[0]
    # which lists a bunch of urls as starting points
    urls = r.json()
    images_url = urls['url:images']
    single_image_url = images_url+str(img_id)+"/"
    surl = session.get(single_image_url)
    try:
        surl.raise_for_status()
    except requests.exceptions.HTTPError as e:
        # Whoops it wasn't a 200
        print("Error {}: received response {} \
                         ".format(e, surl.status_code))
        raise
    # calculate width to be requested
    thisjson = surl.json()
    width = int(thisjson['data']['Pixels']['SizeX'])
    scaled = round(width/scale)
    img_address = host + "/webgateway/render_birds_eye_view/" + \
        str(img_id)+"/"+str(scaled)+"/"
    jpeg = session.get(img_address, stream=True)
    try:
        jpeg.raise_for_status()
    except requests.exceptions.HTTPError as e:
        # Whoops it wasn't a 200
        print("Error {!r}: received response {!r} with content: \
                         {!r}".format(e, jpeg.status_code, jpeg.content))
        raise

    # using PIL and BytesIO to open the request content as an image
    i = Image.open(BytesIO(jpeg.content))
    jpeg.close()
    return np.array(i)



import os
import pytest
import subprocess
import numpy as np
from datetime import datetime
import ezomero
from ezomero import rois
from omero.cli import CLI
from omero.gateway import BlitzGateway
from omero.gateway import ScreenWrapper, PlateWrapper
from omero.model import ScreenI, PlateI, WellI, WellSampleI, ImageI
from omero.model import ScreenPlateLinkI, PlateAcquisitionI
from omero.plugins.sessions import SessionsControl
from omero.plugins.user import UserControl
from omero.plugins.group import GroupControl
from omero.rtypes import rint
import importlib.util
# try importing pandas
if (importlib.util.find_spec('pandas')):
    import pandas as pd
    has_pandas = True
else:
    has_pandas = False


# Settings for OMERO
DEFAULT_OMERO_USER = "root"
DEFAULT_OMERO_PASS = "omero"
DEFAULT_OMERO_HOST = "localhost"
DEFAULT_OMERO_WEB_HOST = "http://localhost:5080"
DEFAULT_OMERO_PORT = 6064
DEFAULT_OMERO_SECURE = 1

# [[group, permissions], ...]
GROUPS_TO_CREATE = [['test_group_1', 'read-only'],
                    ['test_group_2', 'read-only']]

# [[user, [groups to be added to], [groups to own]], ...]
USERS_TO_CREATE = [
                   [
                    'test_user1',
                    ['test_group_1', 'test_group_2'],
                    ['test_group_1']
                   ],
                   [
                    'test_user2',
                    ['test_group_1', 'test_group_2'],
                    ['test_group_2']
                   ],
                   [
                    'test_user3',
                    ['test_group_2'],
                    []
                   ]
                  ]


def pytest_addoption(parser):
    parser.addoption("--omero-user",
                     action="store",
                     default=os.environ.get("OMERO_USER",
                                            DEFAULT_OMERO_USER))
    parser.addoption("--omero-pass",
                     action="store",
                     default=os.environ.get("OMERO_PASS",
                                            DEFAULT_OMERO_PASS))
    parser.addoption("--omero-host",
                     action="store",
                     default=os.environ.get("OMERO_HOST",
                                            DEFAULT_OMERO_HOST))
    parser.addoption("--omero-web-host",
                     action="store",
                     default=os.environ.get("OMERO_WEB_HOST",
                                            DEFAULT_OMERO_WEB_HOST))
    parser.addoption("--omero-port",
                     action="store",
                     type=int,
                     default=int(os.environ.get("OMERO_PORT",
                                                DEFAULT_OMERO_PORT)))
    parser.addoption("--omero-secure",
                     action="store",
                     default=bool(os.environ.get("OMERO_SECURE",
                                                 DEFAULT_OMERO_SECURE)))


# we can change this later
@pytest.fixture(scope="session")
def omero_params(request):
    user = request.config.getoption("--omero-user")
    password = request.config.getoption("--omero-pass")
    host = request.config.getoption("--omero-host")
    web_host = request.config.getoption("--omero-web-host")
    port = request.config.getoption("--omero-port")
    secure = request.config.getoption("--omero-secure")
    return (user, password, host, web_host, port, secure)


@pytest.fixture(scope='session')
def users_groups(conn, omero_params):
    session_uuid = conn.getSession().getUuid().val
    user = omero_params[0]
    host = omero_params[2]
    port = str(omero_params[4])
    cli = CLI()
    cli.register('sessions', SessionsControl, 'test')
    cli.register('user', UserControl, 'test')
    cli.register('group', GroupControl, 'test')

    group_info = []
    for gname, gperms in GROUPS_TO_CREATE:
        cli.invoke(['group', 'add',
                    gname,
                    '--type', gperms,
                    '-k', session_uuid,
                    '-u', user,
                    '-s', host,
                    '-p', port])
        gid = ezomero.get_group_id(conn, gname)
        group_info.append([gname, gid])

    user_info = []
    for user, groups_add, groups_own in USERS_TO_CREATE:
        # make user while adding to first group
        cli.invoke(['user', 'add',
                    user,
                    'test',
                    'tester',
                    '--group-name', groups_add[0],
                    '-e', 'useremail@jax.org',
                    '-P', 'abc123',
                    '-k', session_uuid,
                    '-u', user,
                    '-s', host,
                    '-p', port])

        # add user to rest of groups
        if len(groups_add) > 1:
            for group in groups_add[1:]:
                cli.invoke(['group', 'adduser',
                            '--user-name', user,
                            '--name', group,
                            '-k', session_uuid,
                            '-u', user,
                            '-s', host,
                            '-p', port])

        # make user owner of listed groups
        if len(groups_own) > 0:
            for group in groups_own:
                cli.invoke(['group', 'adduser',
                            '--user-name', user,
                            '--name', group,
                            '--as-owner',
                            '-k', session_uuid,
                            '-u', user,
                            '-s', host,
                            '-p', port])
        uid = ezomero.get_user_id(conn, user)
        user_info.append([user, uid])

    return (group_info, user_info)


@pytest.fixture(scope='session')
def conn(omero_params):
    user, password, host, web_host, port, secure = omero_params
    conn = BlitzGateway(user, password, host=host, port=port, secure=secure)
    conn.connect()
    yield conn
    conn.close()


@pytest.fixture(scope='session')
def image_fixture():
    test_image = np.zeros((200, 201, 20, 3, 1), dtype=np.uint8)
    test_image[0:100, 0:100, 0:10, 0, :] = 255
    test_image[0:100, 0:100, 11:20, 1, :] = 255
    test_image[101:200, 101:201, :, 2, :] = 255
    return test_image


@pytest.fixture(scope='session')
def pyramid_fixture(conn, omero_params):
    session_uuid = conn.getSession().getUuid().val
    user = omero_params[0]
    host = omero_params[2]
    port = str(omero_params[4])
    imp_cmd = ['omero', 'import', 'tests/data/test_pyramid.ome.tif',
               '-k', session_uuid,
               '-u', user,
               '-s', host,
               '-p', port]
    process = subprocess.Popen(imp_cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    stdoutval, stderrval = process.communicate()


@pytest.fixture(scope='session')
def roi_fixture():
    point = rois.Point(x=100.0, y=100.0, z=0, c=0, t=0, label='test_point',
                       fill_color=(0, 1, 2, 3), stroke_color=(4, 5, 6, 100),
                       stroke_width=1.0)
    line = rois.Line(x1=100.0, y1=100.0, x2=150.0, y2=150.0, z=0, c=0, t=0,
                     label='test_line', fill_color=(7, 8, 9, 10),
                     stroke_color=(11, 12, 13, 106), stroke_width=2.0)
    arrow = rois.Line(x1=100.0, y1=100.0, x2=150.0, y2=150.0, z=0, c=0, t=0,
                      label='test_arrow', markerEnd="Arrow",
                      markerStart="Arrow")
    rectangle = rois.Rectangle(x=100.0, y=100.0, width=50.0, height=40.0, z=0,
                               c=0, t=0, label='test_rectangle',
                               fill_color=(14, 15, 16, 17),
                               stroke_color=(18, 19, 20, 101),
                               stroke_width=3.0)
    ellipse = rois.Ellipse(x=80, y=60, x_rad=20.0, y_rad=40.0, z=0, c=0, t=0,
                           label='test_ellipse',
                           fill_color=(21, 22, 23, 24),
                           stroke_color=(25, 26, 27, 102),
                           stroke_width=4.0)
    polygon = rois.Polygon(points=[(100.0, 100.0),
                                   (110.0, 150.0),
                                   (100.0, 150.0)],
                           z=0, c=0, t=0, label='test_polygon',
                           fill_color=(28, 29, 30, 31),
                           stroke_color=(32, 33, 34, 103),
                           stroke_width=5.0)
    polyline = rois.Polyline(points=[(100.0, 100.0),
                                     (110.0, 150.0),
                                     (100.0, 150.0)],
                             z=0, c=0, t=0, label='test_polyline',
                             fill_color=(35, 36, 37, 38),
                             stroke_color=(39, 40, 41, 104),
                             stroke_width=6.0)
    label = rois.Label(x=100.0, y=100.0, z=0, c=0, t=0,
                       label='test_label', fontSize=60,
                       fill_color=(42, 43, 44, 45),
                       stroke_color=(46, 47, 48, 105),
                       stroke_width=7.0)

    return {'shapes': [point, line, rectangle, ellipse,
                       polygon, polyline, arrow, label],
            'name': 'ROI_name',
            'desc': 'A description for the ROI'
            }


@pytest.fixture(scope='session')
def timestamp():
    return f'{datetime.now():%Y%m%d%H%M%S}'


@pytest.fixture(scope='session')
def project_structure(conn, timestamp, image_fixture, users_groups,
                      omero_params):
    group_info, user_info = users_groups
    # Don't change anything for default_user!
    # If you change anything about users/groups, make sure they exist
    # [[group, [projects]], ...] per user
    project_str = {
                    'users': [
                        {
                            'name': 'default_user',
                            'groups': [
                                {
                                    'name': 'default_group',
                                    'projects': [
                                        {
                                            'name': f'proj0_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds0_{timestamp}',
                                                    'images': [
                                                        f'im0_{timestamp}'
                                                    ]
                                                }
                                            ]
                                        }
                                    ],
                                    'datasets': []
                                }
                            ]
                        },
                        {
                            'name': 'test_user1',
                            'groups': [
                                {
                                    'name': 'test_group_1',
                                    'projects': [
                                        {
                                            'name': f'proj1_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds1_{timestamp}',
                                                    'images': [
                                                        f'im1_{timestamp}'
                                                    ]
                                                }
                                            ]
                                        },
                                        {
                                            'name': f'proj2_{timestamp}',
                                            'datasets': []
                                        }
                                    ],
                                    'datasets': []
                                },
                                {
                                    'name': 'test_group_2',
                                    'projects': [
                                        {
                                            'name': f'proj3_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds2_{timestamp}',
                                                    'images': [
                                                        f'im2_{timestamp}'

                                                    ]
                                                },
                                                {
                                                    'name': f'ds3_{timestamp}',
                                                    'images': [
                                                        f'im3_{timestamp}',
                                                        f'im4_{timestamp}'
                                                    ]
                                                }
                                            ]
                                        }
                                    ],
                                    'datasets': []
                                }
                            ]
                        },
                        {
                            'name': 'test_user2',
                            'groups': [
                                {
                                    'name': 'test_group_1',
                                    'projects': [
                                        {
                                            'name': f'proj4_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds4_{timestamp}',
                                                    'images': [
                                                        f'im5_{timestamp}'
                                                    ]
                                                }
                                            ]
                                        },
                                        {
                                            'name': f'proj5_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds5_{timestamp}',
                                                    'images': [

                                                    ]
                                                }
                                            ]
                                        }
                                    ],
                                    'datasets': []
                                },
                                {
                                    'name': 'test_group_2',
                                    'projects': [
                                        {
                                            'name': f'proj6_{timestamp}',
                                            'datasets': [
                                                {
                                                    'name': f'ds6_{timestamp}',
                                                    'images': [
                                                        f'im6_{timestamp}',
                                                        f'im7_{timestamp}'
                                                    ]
                                                }
                                            ]
                                        }
                                    ],
                                    'datasets': [
                                        {
                                            'name': f'ds7_{timestamp}'
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                  }
    project_info = []
    dataset_info = []
    image_info = []
    for user in project_str['users']:
        username = user['name']
        for group in user['groups']:
            groupname = group['name']
            current_conn = conn

            # New connection if user and group need to be specified
            if username != 'default_user':
                current_conn = conn.suConn(username, groupname)

            # Loop to post projects, datasets, and images
            for project in group['projects']:
                projname = project['name']
                proj_id = ezomero.post_project(current_conn,
                                               projname,
                                               'test project')
                project_info.append([projname, proj_id])

                for dataset in project['datasets']:
                    dsname = dataset['name']
                    ds_id = ezomero.post_dataset(current_conn,
                                                 dsname,
                                                 proj_id,
                                                 'test dataset')
                    dataset_info.append([dsname, ds_id])

                    for imname in dataset['images']:
                        im_id = ezomero.post_image(current_conn,
                                                   image_fixture,
                                                   imname,
                                                   dataset_id=ds_id)
                        image_info.append([imname, im_id])
            for dataset in group['datasets']:
                dsname = dataset['name']
                ds_id = ezomero.post_dataset(current_conn,
                                             dsname,
                                             description='test dataset')
                dataset_info.append([dsname, ds_id])

            # Close temporary connection if it was created
            if username != 'default_user':
                current_conn.close()

    yield [project_info, dataset_info, image_info]
    current_group = conn.getGroupFromContext().getId()
    conn.SERVICE_OPTS.setOmeroGroup(-1)
    for dname, did in dataset_info:
        conn.deleteObjects("Dataset", [did], deleteAnns=True,
                           deleteChildren=True, wait=True)
    for pname, pid in project_info:
        conn.deleteObjects("Project", [pid], deleteAnns=True,
                           deleteChildren=True, wait=True)
    conn.SERVICE_OPTS.setOmeroGroup(current_group)


@pytest.fixture(scope='session')
def screen_structure(conn, timestamp, image_fixture):
    # screen info
    update_service = conn.getUpdateService()
    # Create Screen
    screen_name = "screen_" + timestamp
    screen = ScreenWrapper(conn, ScreenI())
    screen.setName(screen_name)
    screen.save()
    screen_id = screen.getId()
    # Create Plate
    plate_name = "plate1_" + timestamp
    plate = PlateWrapper(conn, PlateI())
    plate.setName(plate_name)
    plate.save()
    plate_id = plate.getId()
    link = ScreenPlateLinkI()
    link.setParent(ScreenI(screen_id, False))
    link.setChild(PlateI(plate_id, False))
    update_service.saveObject(link)

    # Create second Plate
    plate2_name = "plate2_" + timestamp
    plate2 = PlateWrapper(conn, PlateI())
    plate2.setName(plate2_name)
    plate2.save()
    plate2_id = plate2.getId()
    link = ScreenPlateLinkI()
    link.setParent(ScreenI(screen_id, False))
    link.setChild(PlateI(plate2_id, False))
    update_service.saveObject(link)

    # Create Well (row 1, col 1)
    well1 = WellI()
    well1.setPlate(PlateI(plate_id, False))
    well1.setColumn(rint(1))
    well1.setRow(rint(1))

    # Create another Well (row 2, col 2)
    well2 = WellI()
    well2.setPlate(PlateI(plate_id, False))
    well2.setColumn(rint(2))
    well2.setRow(rint(2))

    # Create Well for second plate
    well3 = WellI()
    well3.setPlate(PlateI(plate2_id, False))
    well3.setColumn(rint(2))
    well3.setRow(rint(2))

    # Create PlateAcquisition/Run for plate 1
    run1 = PlateAcquisitionI()
    run1.setPlate(PlateI(plate_id, False))
    run2 = PlateAcquisitionI()
    run2.setPlate(PlateI(plate_id, False))

    # Create PlateAcquisition/Run for plate 2
    run3 = PlateAcquisitionI()
    run3.setPlate(PlateI(plate2_id, False))

    well1 = update_service.saveAndReturnObject(well1)
    well2 = update_service.saveAndReturnObject(well2)
    well3 = update_service.saveAndReturnObject(well3)
    well1_id = well1.getId().getValue()
    well2_id = well2.getId().getValue()
    well3_id = well3.getId().getValue()

    run1 = update_service.saveAndReturnObject(run1)
    run2 = update_service.saveAndReturnObject(run2)
    run3 = update_service.saveAndReturnObject(run3)
    run1_id = run1.getId().getValue()
    run2_id = run2.getId().getValue()
    run3_id = run3.getId().getValue()

    # Create Well Sample with Image for both wells
    ws = WellSampleI()
    im_id1 = ezomero.post_image(conn, image_fixture, "well image")
    ws.setImage(ImageI(im_id1, False))
    well1.addWellSample(ws)
    run1.addWellSample(ws)

    ws2 = WellSampleI()
    im_id2 = ezomero.post_image(conn, image_fixture, "well image2")
    ws2.setImage(ImageI(im_id2, False))
    well2.addWellSample(ws2)
    run1.addWellSample(ws2)

    ws3 = WellSampleI()
    im_id3 = ezomero.post_image(conn, image_fixture, "well image3")
    ws3.setImage(ImageI(im_id3, False))
    well1.addWellSample(ws3)
    run2.addWellSample(ws3)

    ws4 = WellSampleI()
    im_id4 = ezomero.post_image(conn, image_fixture, "well image4")
    ws4.setImage(ImageI(im_id4, False))
    well2.addWellSample(ws4)
    run2.addWellSample(ws4)

    ws5 = WellSampleI()
    im_id5 = ezomero.post_image(conn, image_fixture, "well image5")
    ws5.setImage(ImageI(im_id5, False))
    well3.addWellSample(ws5)
    run3.addWellSample(ws5)

    # One call for each plate is enough to update
    well1 = update_service.saveAndReturnObject(well1)
    well3 = update_service.saveAndReturnObject(well3)

    # Create OrphanPlate
    plate3_name = "plate3_" + timestamp
    plate3 = PlateWrapper(conn, PlateI())
    plate3.setName(plate3_name)
    plate3.save()
    plate3_id = plate3.getId()

    # Create Well fr orphan plate
    well4 = WellI()
    well4.setPlate(PlateI(plate3_id, False))
    well4.setColumn(rint(1))
    well4.setRow(rint(1))

    # Create Well Sample with Image
    ws6 = WellSampleI()
    im_id6 = ezomero.post_image(conn, image_fixture, "well image6")
    ws6.setImage(ImageI(im_id6, False))
    well4.addWellSample(ws6)

    well4 = update_service.saveAndReturnObject(well4)
    well4_id = well4.getId().getValue()

    yield [screen_id, plate_id, plate2_id, plate3_id,
           run1_id, run2_id, run3_id,
           well1_id, im_id1, im_id3,
           well2_id, im_id2, im_id4,
           well3_id, im_id5,
           well4_id, im_id6]
    current_group = conn.getGroupFromContext().getId()
    conn.SERVICE_OPTS.setOmeroGroup(-1)
    conn.deleteObjects("Screen", [screen_id], deleteAnns=True,
                       deleteChildren=True, wait=True)
    conn.deleteObjects("Plate", [plate3_id], deleteAnns=True,
                       deleteChildren=True, wait=True)
    conn.SERVICE_OPTS.setOmeroGroup(current_group)


@pytest.fixture(scope='session')
def tables():
    table = [
        ['intcol', 'floatcol', 'stringcol', 'boolcol', 'mixed'],
        [1, 1.2, 'string1', True, 'mixedstr'],
        [2, 2.3, 'string2', False, 1],
        [3, 3.4, 'string3', False, 2.4],
        [4, 4.5, 'string4', True, True],
    ]
    result_table = [
        ['intcol', 'floatcol', 'stringcol', 'boolcol'],
        [1, 1.2, 'string1', True],
        [2, 2.3, 'string2', False],
        [3, 3.4, 'string3', False],
        [4, 4.5, 'string4', True],
    ]
    return [table, result_table]


@pytest.fixture(scope='session')
def table_dfs():
    table = [
        ['intcol', 'floatcol', 'stringcol', 'boolcol', 'mixed'],
        [1, 1.2, 'string1', True, 'mixedstr'],
        [2, 2.3, 'string2', False, 1],
        [3, 3.4, 'string3', False, 2.4],
        [4, 4.5, 'string4', True, True],
    ]
    result_table = [
        ['intcol', 'floatcol', 'stringcol', 'boolcol'],
        [1, 1.2, 'string1', True],
        [2, 2.3, 'string2', False],
        [3, 3.4, 'string3', False],
        [4, 4.5, 'string4', True],
    ]
    headers = table.pop(0)
    df = pd.DataFrame(table, columns=headers)
    headers = result_table.pop(0)
    result_df = pd.DataFrame(result_table, columns=headers)
    return [df, result_df]



import configparser
import logging
import os
import functools
import inspect
from typing import Callable, Optional, Union
from getpass import getpass
from omero.gateway import BlitzGateway
from pathlib import Path


def get_default_args(func: Callable) -> dict:
    """Retrieves the default arguments of a function.

    Parameters
    ----------
    func : function
        Function whose signature we want to inspect

    Returns
    -------
    _ : dict
        Key-value pairs of argument name and value.
    """
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def do_across_groups(f: Callable) -> object:
    """Decorator functional for making functions work across
    OMERO groups.

    Parameters
    ----------
    f : function
        Function that will be decorated

    Returns
    -------
    wrapper : Object
        Return value of the decorated function.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        defaults = get_default_args(f)
        do_across_groups = False
        # test if user is overriding default
        if 'across_groups' in kwargs:
            # if they are, respect user settings
            if kwargs['across_groups']:
                do_across_groups = True
        else:
            # else, respect default
            if defaults['across_groups']:
                do_across_groups = True
        if do_across_groups:
            current_group = args[0].getGroupFromContext().getId()
            args[0].SERVICE_OPTS.setOmeroGroup('-1')
            res = f(*args, **kwargs)
            set_group(args[0], current_group)
        else:
            res = f(*args, **kwargs)
        return res
    return wrapper


# puts
@do_across_groups
def put_map_annotation(conn: BlitzGateway, map_ann_id: int, kv_dict: dict,
                       ns: Union[str, None] = None, across_groups: bool = True
                       ) -> None:
    """Update an existing map annotation with new values (kv pairs)

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    map_ann_id : int
        ID of map annotation whose values (kv pairs) will be replaced.
    kv_dict : dict
        New values (kv pairs) for the MapAnnotation.
    ns : str
        New namespace for the MapAnnotation. If left as None, the old
        namespace will be used.
    across_groups : bool, optional
        Defines cross-group behavior of function - set to
        ``False`` to disable it.

    Notes
    -----
    All keys and values are converted to strings before saving in OMERO.

    Returns
    -------
    Returns None.

    Examples
    --------
    # Change only the values of an existing map annotation:

    >>> new_values = {'key1': 'value1', 'key2': ['value2', 'value3']}
    >>> put_map_annotation(conn, 15, new_values)

    # Change both the values and namespace of an existing map annotation:

    >>> put_map_annotation(conn, 16, new_values, 'test_v2')
    """
    if type(map_ann_id) is not int:
        raise TypeError('Map annotation ID must be an integer')

    map_ann = conn.getObject('MapAnnotation', map_ann_id)
    if map_ann is None:
        raise ValueError("MapAnnotation is non-existent or you do not have "
                         "permissions to change it.")

    if ns is None:
        ns = map_ann.getNs()
    map_ann.setNs(ns)

    kv_pairs = []
    for k, v in kv_dict.items():
        k = str(k)
        if type(v) != list:
            v = str(v)
            kv_pairs.append([k, v])
        else:
            for value in v:
                value = str(value)
                kv_pairs.append([k, value])
                
    map_ann.setValue(kv_pairs)
    map_ann.save()
    return None


@do_across_groups
def put_description(conn: BlitzGateway, obj_type: str, obj_id: int, desc: str,
                    across_groups: bool = True) -> None:
    """Update an existing object description with a new value (string)

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    obj_type: str
        The OMERO object type to have its description changed. Valid object
        types are 'Image', 'Dataset', 'Project', 'FileAnnotation',
        'CommentAnnotation', 'MapAnnotation', 'TagAnnotation', 'Plate',
        'Screen' and 'Roi'.
    obj_id : int
        ID of the object whose description will be replaced.
    desc : str
        New description for the object.
    across_groups : bool, optional
        Defines cross-group behavior of function - set to
        ``False`` to disable it.

    Returns
    -------
    Returns None.

    Examples
    --------
    # Change an image description:

    >>> new_desc = "This is a new description"
    >>> put_description(conn, 'Image', 15, new_desc)

    # Change a TagAnnotation description:

    >>> put_description(conn, 'TagAnnotation', 16, 'new tag description')
    """
    VALID_TYPES = ['Image',
                   'Dataset',
                   'Project',
                   'FileAnnotation',
                   'CommentAnnotation',
                   'MapAnnotation',
                   'TagAnnotation',
                   'Plate',
                   'Screen',
                   'Roi',
                   ]
    if type(obj_type) is not str:
        raise TypeError('Object type must be a string')
    if type(obj_id) is not int:
        raise TypeError('Object ID must be an integer')
    if obj_type not in VALID_TYPES:
        raise ValueError('Object type specified is not valid')

    obj = conn.getObject(obj_type, obj_id)
    if obj is None:
        raise ValueError("Object is non-existent or you do not have "
                         "permissions to change it.")

    obj.setDescription(str(desc))
    obj.save()
    return None


# functions for managing connection context and service options.

def connect(user: Optional[str] = None, password: Optional[str] = None,
            group: Optional[str] = None, host: Optional[str] = None,
            port: Optional[int] = None, secure: Optional[bool] = None,
            config_path: Optional[str] = None) -> Optional[BlitzGateway]:
    """Create an OMERO connection

    This function will create an OMERO connection by populating certain
    parameters for ``omero.gateway.BlitzGateway`` initialization by the
    procedure described in the notes below. Note that this function may
    ask for user input, so be cautious if using in the context of a script.

    Finally, don't forget to close the connection ``conn.close()`` when it
    is no longer needed!

    Parameters
    ----------
    user : str, optional
        OMERO username.

    password : str, optional
        OMERO password.

    group : str, optional
        OMERO group.

    host : str, optional
        OMERO.server host.

    port : int, optional
        OMERO port.

    secure : boolean, optional
        Whether to create a secure session.

    config_path : str, optional
        Path to directory containing '.ezomero' file that stores connection
        information. If left as ``None``, defaults to the home directory as
        determined by Python's ``pathlib``.

    Returns
    -------
    conn : ``omero.gateway.BlitzGateway`` object or None
        OMERO connection, if successful. Otherwise an error is logged and
        returns None.

    Notes
    -----
    The procedure for choosing parameters for ``omero.gateway.BlitzGateway``
    initialization is as follows:

    1) Any parameters given to `ezomero.connect` will be used to initialize
       ``omero.gateway.BlitzGateway``

    2) If a parameter is not given to `ezomero.connect`, populate from
       variables in ``os.environ``:
        * OMERO_USER
        * OMERO_PASS
        * OMERO_GROUP
        * OMERO_HOST
        * OMERO_PORT
        * OMERO_SECURE

    3) If environment variables are not set, try to load from a config file.
       This file should be called '.ezomero'. By default, this function will
       look in the home directory, but ``config_path`` can be used to specify
       a directory in which to look for '.ezomero'.

       The function ``ezomero.store_connection_params`` can be used to create
       the '.ezomero' file.

       Note that passwords can not be loaded from the '.ezomero' file. This is
       to discourage storing credentials in a file as cleartext.

    4) If any remaining parameters have not been set by the above steps, the
       user is prompted to enter a value for each unset parameter.
    """
    if secure and not isinstance(secure, bool):
        raise TypeError("'secure' variable must be a boolean")
    # load from .ezomero config file if it exists
    if config_path is None:
        config_fp = Path.home() / '.ezomero'
    elif type(config_path) is str:
        config_fp = Path(config_path) / '.ezomero'
    else:
        raise TypeError('config_path must be a string')
    config_dict: Union[None, configparser.SectionProxy] = None
    if config_fp.exists():
        config = configparser.ConfigParser()
        with config_fp.open() as fp:
            config.read_file(fp)
        config_dict = config["DEFAULT"]

    # set user
    if user is None:
        if config_dict:
            user = config_dict.get("OMERO_USER", user)
        user = os.environ.get("OMERO_USER", user)
    if user is None:
        user = input('Enter username: ')

    # set password
    if password is None:
        password = os.environ.get("OMERO_PASS", password)
    if password is None:
        password = getpass('Enter password: ')

    # set group
    if group is None:
        if config_dict:
            group = config_dict.get("OMERO_GROUP", group)
        group = os.environ.get("OMERO_GROUP", group)
    if group is None:
        group = input('Enter group name (or leave blank for default group): ')
    if group == "":
        group = None

    # set host
    if host is None:
        if config_dict:
            host = config_dict.get("OMERO_HOST", host)
        host = os.environ.get("OMERO_HOST", host)
    if host is None:
        host = input('Enter host: ')

    # set port
    port_str = None
    if port is None:
        if config_dict:
            port_str = config_dict.get("OMERO_PORT", str(port))
        port_str = os.environ.get("OMERO_PORT", port_str)
        if port_str:
            port = int(port_str)
    if port is None:
        port = int(input('Enter port: '))

    # set session security
    secure_str = None
    if secure is None:
        if config_dict:
            secure_str = config_dict.get("OMERO_SECURE", str(secure))
    secure_str = os.environ.get("OMERO_SECURE", secure_str)
    if secure_str:
        if secure_str.lower() in ["true", "t"]:
            secure = True
        elif secure_str.lower() in ["false", "f"]:
            secure = False
    if secure is None:
        str_secure: str = input('Secure session (True or False): ')
        if str_secure.lower() in ["true", "t"]:
            secure = True
        elif str_secure.lower() in ["false", "f"]:
            secure = False
        else:
            raise ValueError('secure must be set to either True or False')
    # create connection
    conn = BlitzGateway(user, password, group=group, host=host, port=port,
                        secure=secure)
    if conn.connect():
        return conn
    else:
        logging.error('Could not connect, check your settings')
        return None


def store_connection_params(user: Optional[str] = None,
                            group: Optional[str] = None,
                            host: Optional[str] = None,
                            port: Optional[int] = None,
                            secure: Optional[bool] = None,
                            web_host: Optional[Union[str, bool]] = False,
                            config_path: Optional[str] = None):
    """Save OMERO connection parameters in a file.

    This function creates a config file ('.ezomero') in which
    certain OMERO parameters are stored, to make it easier to create
    ``omero.gateway.BlitzGateway`` objects.

    Parameters
    ----------
    user : str, optional
        OMERO username.

    group : str, optional
        OMERO group.

    host : str, optional
        OMERO.server host.

    port : int, optional
        OMERO port.

    secure : boolean, optional
        Whether to create a secure session.

    web_host : boolean/str, optional
        Whether to save a web host address got JSON connections as well. If
        `False`, will skip it; it `True`, will prompt user for it; if it is
        a `str`, will save that value to `OMERO_WEB_HOST`.

    config_path : str, optional
        Path to directory that will contain the '.ezomero' file. If left as
        ``None``, defaults to the home directory as determined by Python's
        ``pathlib``.
    """
    if config_path is None:
        cpath = Path.home()
    elif type(config_path) is str:
        cpath = Path(config_path)
    else:
        raise ValueError('config_path must be a string')

    if not cpath.is_dir():
        raise ValueError('config_path must point to a valid directory')
    ezo_file = cpath / '.ezomero'
    if ezo_file.exists():
        resp = input(f'{ezo_file} already exists. Overwrite? (Y/N)')
        if resp.lower() not in ['yes', 'y']:
            return

    # get parameters
    if user is None:
        user = input('Enter username: ')
    if group is None:
        group = input('Enter group name (or leave blank for default group): ')
    if host is None:
        host = input('Enter host: ')
    if port is None:
        port = int(input('Enter port: '))
    if secure is None:
        secure_str = input('Secure session (True or False): ')
        if secure_str.lower() in ["true", "t"]:
            secure = True
        elif secure_str.lower() in ["false", "f"]:
            secure = False
        else:
            raise ValueError('secure must be set to either True or False')
    if web_host is True:
        web_host_str = input('Enter web host: ')
    # make parameter dictionary and save as configfile
    # just use 'DEFAULT' for right now, we can possibly add alt configs later
    config = configparser.ConfigParser()
    config['DEFAULT'] = {'OMERO_USER': user,
                         'OMERO_GROUP': group,
                         'OMERO_HOST': host,
                         'OMERO_PORT': str(port),
                         'OMERO_SECURE': str(secure)}
    if web_host is not False:
        if isinstance(web_host, str):
            web_host_str = web_host
        config['JSON'] = {'OMERO_USER': user,
                          'OMERO_WEB_HOST': web_host_str,
                          }
    with ezo_file.open('w') as configfile:
        config.write(configfile)
        print(f'Connection settings saved to {ezo_file}')


def set_group(conn: BlitzGateway, group_id: int) -> bool:
    """Safely switch OMERO group.

    This function will change the user's current group to that specified by
    `group_id`, but only if the user is a member of that group. This is a
    "safer" way to do this than ``conn.SERVICE_OPTS.setOmeroGroup``, which will
    allow switching to a group that a user does not have permissions, which can
    lead to server-side errors.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    group_id : int
        The id of the group to switch to.

    Returns
    -------
    change_status : bool
        Returns `True` if group is changed, otherwise returns `False`.
    """
    if type(group_id) is not int:
        raise TypeError('Group ID must be an integer')

    user_id = conn.getUser().getId()
    g = conn.getObject("ExperimenterGroup", group_id)
    owners, members = g.groupSummary()
    owner_ids = [e.getId() for e in owners]
    member_ids = [e.getId() for e in members]
    if (user_id in owner_ids) or (user_id in member_ids):
        conn.SERVICE_OPTS.setOmeroGroup(group_id)
        return True
    else:
        logging.warning(f'User {user_id} is not a member of Group {group_id}')
        return False



from ._ezomero import do_across_groups
from typing import Optional, List
from omero.sys import Parameters
from omero.gateway import BlitzGateway
from omero.rtypes import rstring
from omero.model import DatasetImageLinkI, ImageI, ExperimenterI
from omero.model import DatasetI, ProjectI, ProjectDatasetLinkI
from omero.model import PlateI, ScreenI, ScreenPlateLinkI


# filters
@do_across_groups
def filter_by_filename(conn: BlitzGateway, im_ids: List[int],
                       imported_filename: str,
                       across_groups: Optional[bool] = True) -> List[int]:
    """Filter list of image ids by originalFile name.

    Sometimes we know the filename of an image that has been imported into
    OMERO but not necessarily the image ID. This is frequently the case when
    we want to annotate a recently imported image. This funciton will help
    to filter a list of image IDs to only those associated with a particular
    filename.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    im_ids : list of int
        List of OMERO image IDs.
    imported_filename : str
        The full filename (with extension) of the file whose OMERO image
        we are looking for. NOT the path of the image.
    across_groups : bool, optional
        Defines cross-group behavior of function - set to
        ``False`` to disable it.

    Returns
    -------
    filtered_im_ids : list of int
        Filtered list of images with originalFile name matching
        ``imported_filename``.

    Notes
    -----
    This function should be used as a filter on an image list that has been
    already narrowed down as much as possible. Note that many different images
    in OMERO may share the same filename (e.g., image.tif).

    Examples
    --------
    >>> im_ids = get_image_ids(conn, dataset=303)
    >>> im_ids = filter_by_filename(conn, im_ids, "feb_2020.tif")]
    """
    if not isinstance(im_ids, list):
        raise TypeError('Image IDs must be a list of integers')

    if type(imported_filename) is not str:
        raise TypeError('Imported filename must be a string')

    q = conn.getQueryService()
    params = Parameters()
    params.map = {"oname": rstring(imported_filename)}
    results = q.projection(
        "SELECT i.id FROM Image i"
        " JOIN i.fileset fs"
        " JOIN fs.usedFiles u"
        " JOIN u.originalFile o"
        " WHERE o.name=:oname",
        params,
        conn.SERVICE_OPTS
        )
    im_id_matches = [r[0].val for r in results]

    return list(set(im_ids) & set(im_id_matches))


@do_across_groups
def filter_by_tag_value(conn: BlitzGateway, im_ids: List[int],
                        tag_value: str, across_groups: Optional[bool] = True
                        ) -> List[int]:
    """Filter list of image ids by textValue of a TagAnnotation.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    im_ids : list of int
        List of OMERO image IDs.
    tag_value : str
        Value of tag to filter on.
    across_groups : bool, optional
        Defines cross-group behavior of function - set to
        ``False`` to disable it.

    Returns
    -------
    filtered_im_ids : list of int
    """
    if not isinstance(im_ids, list):
        raise TypeError('Image IDs must be a list of integers')

    if type(tag_value) is not str:
        raise TypeError('Tag value must be a string')

    q = conn.getQueryService()
    params = Parameters()
    params.map = {"tagvalue": rstring(tag_value)}
    results = q.projection(
        "SELECT i.id FROM Image i"
        " JOIN i.annotationLinks al"
        " JOIN al.child a"
        " WHERE a.textValue=:tagvalue"
        " AND TYPE(a)=TagAnnotation",
        params,
        conn.SERVICE_OPTS
        )
    im_id_matches = [r[0].val for r in results]
    return list(set(im_ids) & set(im_id_matches))


@do_across_groups
def filter_by_kv(conn: BlitzGateway, im_ids: List[int], key: str, value: str,
                 across_groups: Optional[bool] = True) -> List[int]:
    """Filter list of image ids by a key-value pair of a MapAnnotation.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    im_ids : list of int
        List of OMERO image IDs.
    key : str
        Key to filter on.
    value : str
        Value to filter on.
    across_groups : bool, optional
        Defines cross-group behavior of function - set to
        ``False`` to disable it.

    Returns
    -------
    filtered_im_ids : list of int
    """
    if not isinstance(im_ids, list):
        raise TypeError('Image IDs must be a list of integers')

    if type(key) is not str:
        raise TypeError('Key must be a string')

    if type(value) is not str:
        raise TypeError('Value must be a string')

    q = conn.getQueryService()
    params = Parameters()
    params.map = {"key": rstring(key),
                  "value": rstring(value)}
    results = q.projection(
        "SELECT i.id FROM Image i"
        " JOIN i.annotationLinks al"
        " JOIN al.child ann"
        " JOIN ann.mapValue as nv"
        " WHERE nv.name = :key"
        " AND nv.value = :value",
        params,
        conn.SERVICE_OPTS
        )

    im_id_matches = [r[0].val for r in results]

    return list(set(im_ids) & set(im_id_matches))


# linking functions
def link_images_to_dataset(conn: BlitzGateway, image_ids: List[int],
                           dataset_id: int):
    """Link images to the specified dataset.

    Nothing is returned by this function.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    im_ids : list of int
        List of OMERO image IDs.
    dataset_id : int
        Id of dataset to which images will be linked.
    """
    if not isinstance(image_ids, list):
        raise TypeError('Image IDs must be a list of integers')

    if type(dataset_id) is not int:
        raise TypeError('Dataset ID must be an integer')

    user_id = _get_current_user(conn)
    for im_id in image_ids:
        link = DatasetImageLinkI()
        link.setParent(DatasetI(dataset_id, False))
        link.setChild(ImageI(im_id, False))
        link.details.owner = ExperimenterI(user_id, False)
        conn.getUpdateService().saveObject(link, conn.SERVICE_OPTS)


def link_datasets_to_project(conn: BlitzGateway, dataset_ids: List[int],
                             project_id: int):
    """Link datasets to the specified project.

    Nothing is returned by this function.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    dataset_ids : list of int
        List of OMERO Dataset Ids.
    project_id : int
        Id of Project to which Datasets will be linked.
    """
    if not isinstance(dataset_ids, list):
        raise TypeError('Dataset IDs must be a list of integers')

    if type(project_id) is not int:
        raise TypeError('Project ID must be an integer')

    user_id = _get_current_user(conn)
    for did in dataset_ids:
        link = ProjectDatasetLinkI()
        link.setParent(ProjectI(project_id, False))
        link.setChild(DatasetI(did, False))
        link.details.owner = ExperimenterI(user_id, False)
        conn.getUpdateService().saveObject(link, conn.SERVICE_OPTS)


def link_plates_to_screen(conn: BlitzGateway, plate_ids: List[int],
                          screen_id: int):
    """Link plates to the specified screen.

    Nothing is returned by this function.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    plate_ids : list of int
        List of OMERO Plate Ids.
    screen_id : int
        Id of Screen to which Plate will be linked.
    """
    if not isinstance(plate_ids, list):
        raise TypeError('Plate IDs must be a list of integers')

    if type(screen_id) is not int:
        raise TypeError('Screen ID must be an integer')

    user_id = _get_current_user(conn)
    for pid in plate_ids:
        link = ScreenPlateLinkI()
        link.setParent(ScreenI(screen_id, False))
        link.setChild(PlateI(pid, False))
        link.details.owner = ExperimenterI(user_id, False)
        conn.getUpdateService().saveObject(link, conn.SERVICE_OPTS)


def _get_current_user(conn: BlitzGateway) -> int:
    userid = conn.SERVICE_OPTS.getOmeroUser()
    if userid is None:
        userid = conn.getUserId()
    return userid


# prints
def print_map_annotation(conn: BlitzGateway, map_ann_id: int):
    """Print some information and value of a map annotation.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    map_ann_id : int
        Id of the MapAnnotation to be displayed.
    """
    if type(map_ann_id) is not int:
        raise TypeError('Map annotation ID must be an integer')

    map_ann = conn.getObject('MapAnnotation', map_ann_id)
    print(f'Map Annotation: {map_ann_id}')
    print(f'Namespace: {map_ann.getNs()}')
    print('Key-Value Pairs:')
    for k, v in map_ann.getValue():
        print(f'\t{k}:\t{v}')


def print_groups(conn: BlitzGateway):
    """Print all Groups with IDs and membership info.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    """
    user_id = conn.getUser().getId()
    print("Groups:")
    for g in conn.listGroups():
        if g.getId() not in [0, 1, 2]:
            owners, members = g.groupSummary()
            owner_ids = [e.getId() for e in owners]
            member_ids = [e.getId() for e in members]
            if user_id in owner_ids:
                group_status = 'owner'
            elif user_id in member_ids:
                group_status = 'member'
            else:
                group_status = ''
            print(f'{g.getName():>25}: {g.getId()}\t{group_status}')


def print_projects(conn: BlitzGateway):
    """Print all available Projects.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    """
    print("Projects:")
    for p in conn.listProjects():
        print(f'\t{p.getName()}:\t{p.getId()}')


def print_datasets(conn: BlitzGateway, project: Optional[int] = None):
    """Print all available Datasets for a given Project.

    Parameters
    ----------
    conn : ``omero.gateway.BlitzGateway`` object
        OMERO connection.
    project : int or `None`, optional
        ID of Project for which to list datasets. If project is `None`,
        orphans are listed.
    """
    if project is not None:
        if type(project) is not int:
            raise TypeError('Project must be an integer')

        p = conn.getObject("Project", project)
        datasets = p.listChildren()
        print(f'Datasets in Project \"{p.getName()}\":')
    else:
        datasets = conn.listOrphans("Dataset")
        print('Orphaned Datsets:')

    for d in datasets:
        print(f"\t{d.getName()}:\t{d.getId()}")



import pytest
import numpy as np
import ezomero
import filecmp
import os
import sys
from unittest import mock


# Test posts
############
def test_post_dataset(conn, project_structure, users_groups, timestamp):

    # testing sanitized inputs
    with pytest.raises(TypeError):
        _ = ezomero.post_dataset(conn, 10)
    with pytest.raises(TypeError):
        _ = ezomero.post_dataset(conn, "testds", description=10)
    with pytest.raises(TypeError):
        _ = ezomero.post_dataset(conn, "testds", project_id='10')

    # Orphaned dataset, with descripion
    ds_test_name = 'test_post_dataset_' + timestamp
    did = ezomero.post_dataset(conn, ds_test_name, description='New test')
    assert conn.getObject("Dataset", did).getName() == ds_test_name
    assert conn.getObject("Dataset", did).getDescription() == "New test"

    # Dataset in default project, no description
    ds_test_name2 = 'test_post_dataset2_' + timestamp
    project_info = project_structure[0]
    pid = project_info[0][1]
    did2 = ezomero.post_dataset(conn, ds_test_name2, project_id=pid)
    ds = conn.getObjects("Dataset", opts={'project': pid})
    ds_names = [d.getName() for d in ds]
    assert ds_test_name2 in ds_names

    # Dataset in non-existing project ID
    ds_test_name3 = 'test_post_dataset3_' + timestamp
    pid = 99999999
    did3 = ezomero.post_dataset(conn, ds_test_name3, project_id=pid)
    ds_ids = ezomero.get_dataset_ids(conn)
    assert did3 is None
    assert len(ds_ids) == 2

    # Dataset in cross-group project, valid permissions
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]   # test_group_1
    current_conn = conn.suConn(username, groupname)
    ds_test_name4 = 'test_post_dataset4_' + timestamp
    project_info = project_structure[0]
    pid = project_info[3][1]  # proj3 (in test_group_2)
    did4 = ezomero.post_dataset(current_conn, ds_test_name4, project_id=pid)
    current_conn.SERVICE_OPTS.setOmeroGroup('-1')
    ds = current_conn.getObjects("Dataset", opts={'project': pid})
    ds_names = [d.getName() for d in ds]
    current_conn.close()
    assert ds_test_name4 in ds_names

    # Dataset in cross-group project, invalid permissions
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    ds_test_name5 = 'test_post_dataset5_' + timestamp
    project_info = project_structure[0]
    pid = project_info[1][1]  # proj1 (in test_group_1)
    did5 = ezomero.post_dataset(current_conn, ds_test_name5, project_id=pid)
    ds_ids = ezomero.get_dataset_ids(current_conn)
    assert len(ds_ids) == 1
    current_conn.close()
    assert did5 is None

    # Dataset in cross-group project, valid permissions
    # across_groups flag unset
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]   # test_group_1
    current_conn = conn.suConn(username, groupname)
    ds_test_name6 = 'test_post_dataset6_' + timestamp
    project_info = project_structure[0]
    pid = project_info[3][1]  # proj3 (in test_group_2)
    did6 = ezomero.post_dataset(current_conn, ds_test_name6, project_id=pid,
                                across_groups=False)
    current_conn.close()
    assert did6 is None

    conn.deleteObjects("Dataset", [did, did2, did4], deleteAnns=True,
                       deleteChildren=True, wait=True)


def test_post_image(conn, project_structure, users_groups, timestamp,
                    image_fixture):
    dataset_info = project_structure[1]
    did = dataset_info[0][1]

    # testing sanitized inputs
    with pytest.raises(TypeError):
        _ = ezomero.post_image(conn, 'image', 'test')
    fake_input = np.zeros((200, 200), dtype=np.uint8)
    with pytest.raises(ValueError):
        _ = ezomero.post_image(conn, fake_input, 'test')
    with pytest.raises(TypeError):
        _ = ezomero.post_image(conn, image_fixture, 10)
    with pytest.raises(TypeError):
        _ = ezomero.post_image(conn, image_fixture, 'test', dataset_id='10')
    with pytest.raises(TypeError):
        _ = ezomero.post_image(conn, image_fixture, 'test', dim_order=10)
    with pytest.raises(ValueError):
        _ = ezomero.post_image(conn, image_fixture, 'test', dim_order='hyzcb')

    # Post image in dataset
    image_name = 'test_post_image_' + timestamp
    im_id = ezomero.post_image(conn, image_fixture, image_name,
                               description='This is an image',
                               dataset_id=did)
    assert conn.getObject("Image", im_id).getName() == image_name

    image_name = 'test_post_image_' + timestamp
    im_id_scr = ezomero.post_image(conn, image_fixture, image_name,
                                   description='This is an image',
                                   dataset_id=did, dim_order='czyxt')
    im = conn.getObject("Image", im_id_scr)
    assert im.getSizeX() == 3
    assert im.getSizeY() == 20
    assert im.getSizeC() == 200

    # Post orphaned image
    im_id2 = ezomero.post_image(conn, image_fixture, image_name)
    assert conn.getObject("Image", im_id2).getName() == image_name

    # Post image to non-existent dataset
    did3 = 999999999
    im_id3 = ezomero.post_image(conn, image_fixture, image_name,
                                description='This is an image',
                                dataset_id=did3)
    assert im_id3 is None

    # Post image cross-group, valid permissions
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    dataset_info = project_structure[1]
    did4 = dataset_info[3][1]  # ds2 (in test_group_2)
    image_name = 'test_post_image_' + timestamp
    im_id4 = ezomero.post_image(current_conn, image_fixture, image_name,
                                description='This is an image',
                                dataset_id=did4)
    current_conn.SERVICE_OPTS.setOmeroGroup('-1')
    assert current_conn.getObject("Image", im_id4).getName() == image_name
    current_conn.close()

    # Post image cross-group, ivvalid permissions
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    dataset_info = project_structure[1]
    did5 = dataset_info[1][1]  # ds1 (in test_group_1)
    image_name = 'test_post_image_' + timestamp
    im_id5 = ezomero.post_image(current_conn, image_fixture, image_name,
                                description='This is an image',
                                dataset_id=did5)
    current_conn.close()
    assert im_id5 is None

    # Post image cross-group, valid permissions, across_groups unset
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    dataset_info = project_structure[1]
    did6 = dataset_info[3][1]  # ds2 (in test_group_2)
    image_name = 'test_post_image_' + timestamp
    im_id6 = ezomero.post_image(current_conn, image_fixture, image_name,
                                description='This is an image',
                                dataset_id=did6, across_groups=False)
    current_conn.close()
    assert im_id6 is None

    conn.deleteObjects("Image", [im_id, im_id2, im_id4, im_id_scr],
                       deleteAnns=True, deleteChildren=True,
                       wait=True)


def test_post_get_map_annotation(conn, project_structure, users_groups):
    image_info = project_structure[2]
    im_id = image_info[0][1]
    # This test both ezomero.post_map_annotation and ezomero.get_map_annotation
    kv = {"key1": "value1",
          "key2": "value2",
          "key3": ["value3", 123]}
    ns = "jax.org/omeroutils/tests/v0"

    # test sanitized input on post
    with pytest.raises(TypeError):
        _ = ezomero.post_map_annotation(conn, "Image", im_id, 'test', ns)
    with pytest.raises(TypeError):
        _ = ezomero.post_map_annotation(conn, "Image", '10', kv, ns)
    with pytest.raises(TypeError):
        _ = ezomero.post_map_annotation(conn, "Image", None, kv, ns)

    map_ann_id = ezomero.post_map_annotation(conn, "Image", im_id, kv, ns)
    kv_pairs = ezomero.get_map_annotation(conn, map_ann_id)
    assert kv_pairs["key2"] == "value2"
    assert sorted(kv_pairs["key3"]) == sorted(["value3", "123"])

    # Test posting to non-existing object
    im_id2 = 999999999
    map_ann_id2 = ezomero.post_map_annotation(conn, "Image", im_id2, kv, ns)
    assert map_ann_id2 is None

    # Test posting cross-group
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id3 = image_info[2][1]  # im2, in test_group_2
    map_ann_id3 = ezomero.post_map_annotation(current_conn, "Image", im_id3,
                                              kv, ns)
    kv_pairs3 = ezomero.get_map_annotation(current_conn, map_ann_id3)
    assert kv_pairs3["key2"] == "value2"
    current_conn.close()

    # Test posting to an invalid cross-group
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    im_id4 = image_info[1][1]  # im1(in test_group_1)
    map_ann_id4 = ezomero.post_map_annotation(current_conn, "Image", im_id4,
                                              kv, ns)
    assert map_ann_id4 is None
    current_conn.close()

    # Test posting cross-group, across_groups unset
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id6 = image_info[2][1]  # im2, in test_group_2
    map_ann_id6 = ezomero.post_map_annotation(current_conn, "Image", im_id6,
                                              kv, ns, across_groups=False)
    assert map_ann_id6 is None
    current_conn.close()

    conn.deleteObjects("Annotation", [map_ann_id, map_ann_id3],
                       deleteAnns=True, deleteChildren=True, wait=True)


def test_post_get_comment_annotation(conn, project_structure, users_groups):
    image_info = project_structure[2]
    im_id = image_info[0][1]
    # This test both ezomero.post_comment_annotation
    # and ezomero.get_comment_annotation
    comment = "sample comment for testing"
    ns = "jax.org/omeroutils/tests/v0"

    # test sanitized input on post
    with pytest.raises(TypeError):
        _ = ezomero.post_comment_annotation(conn, "Image", im_id, 1, ns)
    with pytest.raises(TypeError):
        _ = ezomero.post_comment_annotation(conn, 1, '10', comment, ns)
    with pytest.raises(TypeError):
        _ = ezomero.post_comment_annotation(conn, "Image", None, comment, ns)

    comm_ann_id = ezomero.post_comment_annotation(conn, "Image", im_id,
                                                  comment, ns)
    comm = ezomero.get_comment_annotation(conn, comm_ann_id)
    assert comm == comment

    # Test posting to non-existing object
    im_id2 = 999999999
    comm_ann_id2 = ezomero.post_comment_annotation(conn, "Image", im_id2,
                                                   comment, ns)
    assert comm_ann_id2 is None

    # Test posting cross-group
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id3 = image_info[2][1]  # im2, in test_group_2
    comm_ann_id3 = ezomero.post_comment_annotation(current_conn, "Image",
                                                   im_id3, comment, ns)
    comm = ezomero.get_comment_annotation(current_conn, comm_ann_id3)
    assert comm == comment
    current_conn.close()

    # Test posting to an invalid cross-group
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    im_id4 = image_info[1][1]  # im1(in test_group_1)
    comm_ann_id4 = ezomero.post_comment_annotation(current_conn, "Image",
                                                   im_id4, comment, ns)
    assert comm_ann_id4 is None
    current_conn.close()

    # Test posting cross-group, across_groups unset
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id5 = image_info[2][1]  # im2, in test_group_2
    comm_ann_id5 = ezomero.post_comment_annotation(current_conn, "Image",
                                                   im_id5, comment, ns,
                                                   across_groups=False)
    assert comm_ann_id5 is None
    current_conn.close()

    conn.deleteObjects("Annotation", [comm_ann_id, comm_ann_id3],
                       deleteAnns=True, deleteChildren=True, wait=True)


def test_post_get_file_annotation(conn, project_structure, users_groups,
                                  tmp_path):

    image_info = project_structure[2]
    im_id = image_info[0][1]

    # This test both ezomero.post_file_annotation and
    # ezomero.get_file_annotation
    d = tmp_path / "input"
    d.mkdir()
    file_path = d / "hello.txt"
    file_path.write_text("hello world!")
    file_ann = str(file_path)

    ns = "jax.org/omeroutils/tests/v0"
    # test sanitized input on post
    with pytest.raises(TypeError):
        _ = ezomero.post_file_annotation(conn, 10, ns, "Image", im_id)
    with pytest.raises(TypeError):
        _ = ezomero.post_file_annotation(conn, file_ann, ns, "Image", '10')

    file_ann_id = ezomero.post_file_annotation(conn, file_ann, ns,
                                               "Image", im_id)
    return_ann = ezomero.get_file_annotation(conn, file_ann_id)
    assert filecmp.cmp(return_ann, file_ann)
    os.remove(return_ann)

    # Test posting to non-existing object
    im_id2 = 999999999
    file_ann_id2 = ezomero.post_file_annotation(conn,
                                                file_ann, ns,
                                                "Image", im_id2)
    assert file_ann_id2 is None

    # Test posting cross-group
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id3 = image_info[2][1]  # im2, in test_group_2
    file_ann_id3 = ezomero.post_file_annotation(current_conn,
                                                file_ann, ns,
                                                "Image", im_id3)
    return_ann3 = ezomero.get_file_annotation(current_conn, file_ann_id3)
    assert filecmp.cmp(return_ann3, file_ann)
    os.remove(return_ann3)
    current_conn.close()

    # Test posting to an invalid cross-group
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    im_id4 = image_info[1][1]  # im1(in test_group_1)
    file_ann_id4 = ezomero.post_file_annotation(current_conn,
                                                file_ann, ns,
                                                "Image", im_id4)
    assert file_ann_id4 is None
    current_conn.close()

    # Test posting cross-group, across_groups unset
    username = users_groups[1][0][0]  # test_user1
    groupname = users_groups[0][0][0]  # test_group_1
    current_conn = conn.suConn(username, groupname)
    im_id5 = image_info[2][1]  # im2, in test_group_2
    file_ann_id5 = ezomero.post_file_annotation(current_conn,
                                                file_ann, ns,
                                                "Image", im_id5,
                                                across_groups=False)
    assert file_ann_id5 is None
    current_conn.close()

    # Test posting orphaned
    current_conn = conn.suConn(username, groupname)
    file_ann_id6 = ezomero.post_file_annotation(current_conn, file_ann, ns)
    print(file_ann_id6)
    return_ann6 = ezomero.get_file_annotation(current_conn, file_ann_id6)
    assert filecmp.cmp(return_ann6, file_ann)
    os.remove(return_ann6)

    # Test posting orphaned, partial completion
    file_ann_id7 = ezomero.post_file_annotation(conn, file_ann, ns, "Image")
    return_ann7 = ezomero.get_file_annotation(conn, file_ann_id7)
    assert filecmp.cmp(return_ann7, file_ann)
    os.remove(return_ann7)

    # Test posting orphaned, partial completion
    file_ann_id8 = ezomero.post_file_annotation(conn, file_ann, ns,
                                                object_id=10)
    return_ann8 = ezomero.get_file_annotation(conn, file_ann_id8)
    assert filecmp.cmp(return_ann8, file_ann)
    os.remove(return_ann8)

    conn.deleteObjects("Annotation", [file_ann_id, file_ann_id3, file_ann_id6,
                                      file_ann_id7, file_ann_id8],
                       deleteAnns=True, deleteChildren=True, wait=True)


def test_post_roi(conn, project_structure, roi_fixture, users_groups):
    image_info = project_structure[2]
    im_id = image_info[0][1]

    # test sanitized input on post
    with pytest.raises(TypeError):
        _ = ezomero.post_roi(conn, '10',
                             shapes=roi_fixture['shapes'],
                             name=roi_fixture['name'],
                             description=roi_fixture['desc'])
    with pytest.raises(TypeError):
        _ = ezomero.post_roi(conn, im_id,
                             shapes='10',
                             name=roi_fixture['name'],
                             description=roi_fixture['desc'])
    with pytest.raises(TypeError):
        _ = ezomero.post_roi(conn, im_id,
                             shapes=['10'],
                             name=roi_fixture['name'],
                             description=roi_fixture['desc'])
# "regular" test
    roi_id = ezomero.post_roi(conn, im_id,
                              shapes=roi_fixture['shapes'],
                              name=roi_fixture['name'],
                              description=roi_fixture['desc'])
    roi_in_omero = conn.getObject('Roi', roi_id)
    assert roi_in_omero.getName() == roi_fixture['name']
    assert roi_in_omero.getDescription() == roi_fixture['desc']

    # Test posting to a non-existing image
    im_id2 = 999999999
    with pytest.raises(Exception):  # TODO: verify which exception type
        _ = ezomero.post_roi(conn, im_id2,
                             shapes=roi_fixture['shapes'],
                             name=roi_fixture['name'],
                             description=roi_fixture['desc'],
                             fill_color=roi_fixture['fill_color'],
                             stroke_color=roi_fixture['stroke_color'],
                             stroke_width=roi_fixture['stroke_width'])

    # Test posting to an invalid cross-group
    username = users_groups[1][2][0]  # test_user3
    groupname = users_groups[0][1][0]  # test_group_2
    current_conn = conn.suConn(username, groupname)
    im_id4 = image_info[1][1]  # im1(in test_group_1)
    with pytest.raises(Exception):  # TODO: verify which exception type
        _ = ezomero.post_roi(current_conn, im_id4,
                             shapes=roi_fixture['shapes'],
                             name=roi_fixture['name'],
                             description=roi_fixture['desc'],
                             fill_color=roi_fixture['fill_color'],
                             stroke_color=roi_fixture['stroke_color'],
                             stroke_width=roi_fixture['stroke_width'])
    current_conn.close()

    conn.deleteObjects("Roi", [roi_id], deleteAnns=True,
                       deleteChildren=True, wait=True)


def test_post_project(conn, timestamp):
    # No description
    new_proj = "test_post_project_" + timestamp
    pid = ezomero.post_project(conn, new_proj)
    assert conn.getObject("Project", pid).getName() == new_proj

    # With description
    new_proj2 = "test_post_project2_" + timestamp
    desc = "Now with a description"
    pid2 = ezomero.post_project(conn, new_proj2, description=desc)
    assert conn.getObject("Project", pid2).getDescription() == desc
    conn.deleteObjects("Project", [pid, pid2], deleteAnns=True,
                       deleteChildren=True, wait=True)


def test_post_screen(conn, timestamp):
    # No description
    new_screen = "test_post_screen_" + timestamp
    sid = ezomero.post_screen(conn, new_screen)
    assert conn.getObject("Screen", sid).getName() == new_screen

    # With description
    new_screen2 = "test_post_screen2_" + timestamp
    desc = "Now with a description"
    sid2 = ezomero.post_screen(conn, new_screen2, description=desc)
    assert conn.getObject("Screen", sid2).getDescription() == desc
    conn.deleteObjects("Screen", [sid, sid2], deleteAnns=True,
                       deleteChildren=True, wait=True)


def test_post_project_type(conn):
    with pytest.raises(TypeError):
        _ = ezomero.post_project(conn, 123)
    with pytest.raises(TypeError):
        _ = ezomero.post_project(conn, '123', description=1245)


def test_post_screen_type(conn):
    with pytest.raises(TypeError):
        _ = ezomero.post_screen(conn, 123)
    with pytest.raises(TypeError):
        _ = ezomero.post_screen(conn, '123', description=1245)


def test_post_get_table_pandas(conn, project_structure, table_dfs):

    image_info = project_structure[2]
    im_id = image_info[0][1]

    # This test both ezomero.post_table and ezomero.get_table, using pandas

    table_id = ezomero.post_table(conn, table_dfs[0], "Image", im_id)
    print(table_id)
    return_ann = ezomero.get_table(conn, table_id)
    print(return_ann, table_dfs[1])
    assert return_ann.equals(table_dfs[1])
    conn.deleteObjects("Annotation", [table_id],
                       deleteAnns=True, deleteChildren=True, wait=True)


def test_post_get_table_nopandas(conn, project_structure, users_groups,
                                 tables):

    # This test both ezomero.post_table and ezomero.get_table, not using pandas
    image_info = project_structure[2]
    im_id = image_info[0][1]
    # test sanitized input on post
    with pytest.raises(TypeError):
        _ = ezomero.post_table(conn, 10, "Image", im_id)
    with pytest.raises(TypeError):
        _ = ezomero.post_table(conn, tables[0], "Image", '10')
    with pytest.raises(TypeError):
        _ = ezomero.post_table(conn, tables[0], "Image", None)

    with mock.patch.dict(sys.modules):
        sys.modules["pandas"] = None
        table_id = ezomero.post_table(conn, tables[0], "Image", im_id,
                                      title="test table", headers=True)
        print(table_id)
        return_ann = ezomero.get_table(conn, table_id)
        print(return_ann, tables[1])
        assert return_ann == tables[1]

        # Test posting to non-existing object
        im_id2 = 999999999
        table_id2 = ezomero.post_table(conn, tables[0], "Image", im_id2)
        assert table_id2 is None

        conn.deleteObjects("Annotation", [table_id],
                           deleteAnns=True, deleteChildren=True, wait=True)
